@@ -11,13 +11,13 @@ import unittest
 
 import six
 
-from xpython.pyvm2 import VirtualMachine, VMError
+from xpython.vm import PyVM, PyVMError
 
-from xdis import PYTHON_VERSION
+from xdis import PYTHON_VERSION, IS_PYPY
 
 # Make this false if you need to run the debugger inside a test.
 CAPTURE_STDOUT = ('-s' not in sys.argv)
-# Make this false to see the traceback from a failure inside pyvm2.
+# Make this false to see the traceback from a failure inside pyvm.
 CAPTURE_EXCEPTION = 1
 
 
@@ -44,7 +44,8 @@ class VmTestCase(unittest.TestCase):
             assert PYTHON_VERSION == 2.7
             self.version = 2.7
 
-        path = osp.join(srcdir, "bytecode-%s" % self.version, parent_function_name() + ".pyc")
+        platform='pypy' if IS_PYPY else ''
+        path = osp.join(srcdir, "bytecode-%s%s" % (platform, self.version), parent_function_name() + ".pyc")
         self.assert_ok(path, arg_type="bytecode-file")
 
     def self_checking(self):
@@ -80,12 +81,12 @@ class VmTestCase(unittest.TestCase):
         vm_stdout = six.StringIO()
         if CAPTURE_STDOUT:              # pragma: no branch
             sys.stdout = vm_stdout
-        vm = VirtualMachine(vmtest_testing=True)
+        vm = PyVM(vmtest_testing=True)
 
         vm_value = vm_exc = None
         try:
             vm_value = vm.run_code(code)
-        except VMError:         # pragma: no cover
+        except PyVMError:         # pragma: no cover
             # If the VM code raises an error, show it.
             raise
         except AssertionError:              # pragma: no cover
@@ -151,12 +152,12 @@ class VmTestCase(unittest.TestCase):
         print("%s bytecode %s for %s %s "
               % (LINE_STR, self.version, code.co_filename, LINE_STR))
 
-        vm = VirtualMachine(python_version=self.version)
+        vm = PyVM(python_version=self.version)
 
         vm_value = vm_exc = None
         try:
             vm_value = vm.run_code(code)
-        except VMError:         # pragma: no cover
+        except PyVMError:         # pragma: no cover
             # If the VM code raises an error, show it.
             self.assertTrue(False)
             raise
