@@ -286,14 +286,13 @@ return next yielded value or raise StopIteration."""
         if self.frame is not None:
             block = self.frame.lastblock
             while block is not None:
-                if not isinstance(block, LoopBlock):
+                if block.type == "loop":
                     self.descr_close()
                     break
                 block = block.previous
 
     def frame_is_finished(self):
         self.frame = None
-        rgc.may_ignore_finalizer(self)
 
     def iterator_greenkey(self, space):
         return self.pycode
