@@ -7,7 +7,12 @@ import linecache
 import types
 from copy import copy
 from sys import stderr
-from xdis import CO_GENERATOR, iscode, PYTHON3, PYTHON_VERSION
+from xdis import CO_GENERATOR, CO_ITERABLE_COROUTINE, iscode, PYTHON3, PYTHON_VERSION
+if PYTHON_VERSION >= 3.4:
+    from xpython.stdlib.types34 import _AsyncGeneratorWrapper
+else:
+    class _AsyncGeneratorWrapper():
+        pass
 
 import xpython.stdlib.inspect3 as inspect3
 import xpython.stdlib.inspect2 as inspect2
@@ -42,12 +47,10 @@ except:
 
 
 # Code with these names have an implicit .0 in them
-COMPREHENSION_FN_NAMES = frozenset((
-    "<setcomp>",
-    "<dictcomp>",
-    "<listcomp>",
-    "<genexpr>",
-))
+COMPREHENSION_FN_NAMES = frozenset(
+    ("<setcomp>", "<dictcomp>", "<listcomp>", "<genexpr>",)
+)
+
 
 class Function(object):
     """Function(name, code, globals, argdefs, closure, vm,  kwdefaults={},
@@ -144,7 +147,7 @@ class Function(object):
         self.__dict__ = {
             "version": vm.version,
             "_vm": vm,
-            }
+        }
 
         self.__doc__ = (
             code.co_consts[0] if hasattr(code, "co_consts") and code.co_consts else None
@@ -154,7 +157,7 @@ class Function(object):
             self.__annotations__ = annotations
             self.__kwdefaults__ = kwdefaults
             if vm.version >= 3.4:
-                self.__qualname___ = qualname if qualname else self.__name__
+                self.__qualname__ = qualname if qualname else self.__name__
             else:
                 assert qualname is None
         else:
@@ -240,19 +243,39 @@ class Function(object):
             # It will require a *lot* more code from inspect.py to be added:
             # classes Signature, Parameter, etc.
             callargs = inspect.getcallargs(self._func, *args, **kwargs)
+
+            # The problem with the above is that we are testing with self._func
+            # the function may have changed dynamically.
+            # See 3.7.7. test_keywordonlyarg.py
+
+            # To catch dynamic changes, we'll run a second check
+            if self.version >= 3.0:
+                inspect3.getcallargs(self, *args, **kwargs)
+            else:
+                inspect2.getcallargs(self, *args, **kwargs)
         else:
             if self.version >= 3.0:
                 callargs = inspect3.getcallargs(self, *args, **kwargs)
             else:
                 callargs = inspect2.getcallargs(self, *args, **kwargs)
 
-        frame = self._vm.make_frame(self.func_code, callargs, self.func_globals, {})
-        if self.func_code.co_flags & CO_GENERATOR:
-            gen = Generator(frame, self._vm)
+        frame = self._vm.make_frame(
+            self.func_code, callargs, self.func_globals, {}, self.__closure__
+        )
+        if self.__code__.co_flags & CO_GENERATOR:
+            qualname = self.__qualname__ if self._vm.version >= 3.4 else None
+            gen = Generator(
+                g_frame=frame, name=self.__name__, qualname=qualname, vm=self._vm
+            )
+            if self.__code__.co_flags & CO_ITERABLE_COROUTINE:
+                gen = _AsyncGeneratorWrapper(gen)
+                frame.generator = gen
+                return gen
+
             frame.generator = gen
             retval = gen
         else:
-            retval = self._vm.run_frame(frame)
+            retval = self._vm.eval_frame(frame)
         return retval
 
 
@@ -266,6 +289,14 @@ class Method(object):
         self.im_self = obj
         self.im_class = _class
         self.im_func = func
+
+        if hasattr(func, "func_code"):
+            self.func_code = func.func_code
+        if hasattr(func, "__name__"):
+            self.__name__ = func.__name__
+        if hasattr(func, "__code__"):
+            self.__code__ = func.__code__
+
         # This causes failure
         # self.__name__ = obj.__name__
 
@@ -338,19 +369,19 @@ class Block(object):
 
     def __repr__(self):
         if self.handler is None:
-            return (
-                "<Block type: %s, stack level: %d" %
-                (self.type, self.level)
-            )
+            return "<Block type: %s, stack level: %d" % (self.type, self.level)
         else:
-            return (
-                "<Block type: %s, end offset: @%d, stack level: %d" %
-                (self.type, self.handler, self.level)
+            return "<Block type: %s, end offset: @%d, stack level: %d" % (
+                self.type,
+                self.handler,
+                self.level,
             )
 
 
 class Frame(object):
-    def __init__(self, f_code, f_globals, f_locals, f_back, version=PYTHON_VERSION):
+    def __init__(
+        self, f_code, f_globals, f_locals, f_back, version=PYTHON_VERSION, closure=None
+    ):
         self.f_code = f_code
         self.f_globals = f_globals
         self.f_locals = f_locals
@@ -358,6 +389,11 @@ class Frame(object):
         self.stack = []
         self.f_trace = None
 
+        # brkpt is a mapping bytecode offset to the opcode value that was
+        # smasshed by overwriting it with the pseudo opcode BRKPT.
+        # After a breakpoint is serviced, this opcode needs to be run.
+        self.brkpt = {}
+
         if f_back and f_back.f_globals is f_globals:
             # If we share the globals, we share the builtins.
             self.f_builtins = f_back.f_builtins
@@ -391,17 +427,30 @@ class Frame(object):
         if f_code.co_freevars:
             if not self.cells:
                 self.cells = {}
-            for var in f_code.co_freevars:
-                assert self.cells is not None
-                assert f_back.cells, "f_back.cells: %r" % (f_back.cells,)
-                self.cells[var] = f_back.cells[var]
+            for i, var in enumerate(f_code.co_freevars):
+                if closure:
+                    # print("XXX", f_code.co_freevars[i], closure[i].get())
+                    # if f_code.co_freevars[i] == "c" and  closure[i].get() == 5:
+                    #     from trepan.api import debug; debug()
+                    self.cells[var] = closure[i]
+                else:
+                    # FIXME: this branch is probably wrong.
+                    # Also check all calls of Frame and make_frame() in vm to ensure we pass a function's
+                    # closure attribute.
+                    assert isinstance(f_back.cells[i], Cell), "f_back.cells[%d]: %r" % (
+                        i,
+                        f_back.cells[i],
+                    )
+                    self.cells[var] = f_back.cells[var]
+                pass
+            pass
 
         self.block_stack = []
         self.generator = None
         self.version = version
 
         # These are sentinal or bogus values to start out.
-        # run_frame will adjust inst_index.
+        # eval_frame will adjust inst_index.
         self.inst_index = -1
         self.fallthrough = False
         self.last_op = None
@@ -476,11 +525,16 @@ def traceback_from_frame(frame):
 
 
 class Generator(object):
-    def __init__(self, g_frame, vm):
+    def __init__(self, g_frame, name, qualname, vm):
         self.gi_frame = g_frame
         self.vm = vm
+        self.name = vm
         self.started = False
         self.finished = False
+        self.gi_running = False
+        self.gi_code = g_frame.f_code
+        self.__name__ = g_frame.f_code.co_name
+        self.__qualname__ = qualname if g_frame.version >= 3.4 else None
 
     def __iter__(self):
         return self
@@ -493,8 +547,10 @@ class Generator(object):
             raise TypeError("Can't send non-None value to a just-started generator")
         self.gi_frame.stack.append(value)
         self.started = True
+        self.running = True
         val = self.vm.resume_frame(self.gi_frame)
         if self.finished:
+            self.running = False
             raise StopIteration(val)
         return val
 
@@ -502,14 +558,14 @@ class Generator(object):
 
 
 if __name__ == "__main__":
-    # frame = Frame(
-    #     traceback_from_frame.__code__,
-    #     globals(),
-    #     locals(),
-    #     None,
-    #     PYTHON_VERSION
-    # )
-    # print(frame)
+    frame = Frame(
+        traceback_from_frame.__code__,
+        globals(),
+        locals(),
+        None,
+        PYTHON_VERSION
+    )
+    print(frame)
 
     class Foo(object):
         "Class Foo docstring"
