@@ -4,9 +4,9 @@
 # This file should define a variable VERSION which we use as the
 # debugger version number.
 
-VERSION='1.3.0'  # noqa
+VERSION="1.3.5"  # noqa
 
 # The below has extra commas for POSIX shell's taste, but that's okay
 SUPPORTED_PYTHON=(2.7, 3.3, 3.2, 3.4, 3.5, 3.6, 3.7)  # noqa
 SUPPORTED_PYPY=(2.7, 3.2, 3.5, 3.6)  # noqa
-SUPPORTED_BYTECODE=(2.5, 2.6, 2.7, 3.3, 3.2, 3.4, 3.5, 3.6, 3.7)  # noqa
+SUPPORTED_BYTECODE=(2.4, 2.5, 2.6, 2.7, 3.3, 3.2, 3.4, 3.5, 3.6, 3.7)  # noqa
