@@ -5,7 +5,6 @@
 from __future__ import print_function, division
 import linecache
 import logging
-import operator
 import sys
 
 import six
@@ -22,7 +21,7 @@ from xdis import (
 )
 from xdis.op_imports import get_opcode_module
 
-from xpython.pyobj import Frame, Block, traceback_from_frame
+from xpython.pyobj import Frame, Block, Traceback, traceback_from_frame
 from xpython.byteop import get_byteop
 
 PY2 = not PYTHON3
@@ -62,16 +61,35 @@ class PyVMUncaughtException(Exception):
 
 
 def format_instruction(
-    frame, opc, byteName, intArg, arguments, opoffset, line_number, extra_debug
+    frame, opc, byte_name, int_arg, arguments, offset, line_number, extra_debug, vm=None
 ):
+    """Formats an instruction. What's a little different here is that in
+    contast to Python's `dis`, or a colorized version of that, used in
+    `trepan3k` we may have access to the frame eval stack and therefore
+    can show operands in a nicer way.
+
+    But we also make use of xdis' nicer argument formatting as well. These appear
+    for example in MAKE_FUNCTION, and CALL_FUNCTION.
+    """
     code = frame.f_code if frame else None
-    byteCode = opc.opmap.get(byteName, 0)
-    if hasattr(opc, "opcode_arg_fmt") and byteName in opc.opcode_arg_fmt:
-        argrepr = opc.opcode_arg_fmt[byteName](intArg)
-    elif intArg is None:
+    byte_code = opc.opmap.get(byte_name, 0)
+
+    if vm and byte_name in vm.byteop.stack_fmt:
+        stack_args = vm.byteop.stack_fmt[byte_name](vm, int_arg)
+    else:
+        stack_args = ""
+
+    if vm and byte_name in vm.byteop.stack_fmt:
+        stack_args = vm.byteop.stack_fmt[byte_name](vm, int_arg, repr)
+    else:
+        stack_args = ""
+
+    if hasattr(opc, "opcode_arg_fmt") and byte_name in opc.opcode_arg_fmt:
+        argrepr = opc.opcode_arg_fmt[byte_name](int_arg)
+    elif int_arg is None:
         argrepr = ""
-    elif byteCode in opc.COMPARE_OPS:
-        argrepr = opc.cmp_op[intArg]
+    elif byte_code in opc.COMPARE_OPS:
+        argrepr = opc.cmp_op[int_arg]
     elif isinstance(arguments, list) and arguments:
         argrepr = arguments[0]
     else:
@@ -82,7 +100,7 @@ def format_instruction(
         if line_number is None
         else LINE_NUMBER_WIDTH_FMT % line_number
     )
-    mess = "%s%3d: %s %s" % (line_str, opoffset, byteName, argrepr)
+    mess = "%s%3d: %s%s %s" % (line_str, offset, byte_name, stack_args, argrepr,)
     if extra_debug and frame:
         mess += " %s in %s:%s" % (code.co_name, code.co_filename, frame.f_lineno)
     return mess
@@ -201,9 +219,12 @@ class PyVM(object):
         self.frame.f_lasti = jump
         self.frame.fallthrough = False
 
-    def make_frame(self, code, callargs={}, f_globals=None, f_locals=None):
+    def make_frame(
+        self, code, callargs={}, f_globals=None, f_locals=None, closure=None
+    ):
         # The callargs default is safe because we never modify the dict.
         # pylint: disable=dangerous-default-value
+
         log.debug(
             "make_frame: code=%r, callargs=%s, f_globals=%r, f_locals=%r",
             code,
@@ -233,7 +254,18 @@ class PyVM(object):
             f_locals = {"__locals__": {}}
 
         f_locals.update(callargs)
-        frame = Frame(code, f_globals, f_locals, self.frame, version=self.version)
+        frame = Frame(
+            f_code=code,
+            f_globals=f_globals,
+            f_locals=f_locals,
+            f_back=self.frame,
+            version=self.version,
+            closure=closure,
+        )
+
+        # THINK ABOUT: should this go into making the frame?
+        frame.linestarts = dict(self.opc.findlinestarts(code, dup_lines=True))
+
         log.debug("%r", frame)
         return frame
 
@@ -275,7 +307,7 @@ class PyVM(object):
         else:
             frame.fallthrough = True
 
-        val = self.run_frame(frame)
+        val = self.eval_frame(frame)
         frame.f_back = None
         return val
 
@@ -284,11 +316,11 @@ class PyVM(object):
     ##############################################
 
     # This is the main entry point
-    def run_code(self, code, f_globals=None, f_locals=None):
+    def run_code(self, code, f_globals=None, f_locals=None, toplevel=True):
         """run code using f_globals and f_locals in our VM"""
         frame = self.make_frame(code, f_globals=f_globals, f_locals=f_locals)
         try:
-            val = self.run_frame(frame)
+            val = self.eval_frame(frame)
         except Exception:
             # Until we get test/vmtest.py under control:
             if self.vmtest_testing:
@@ -305,10 +337,11 @@ class PyVM(object):
             raise PyVMUncaughtException
 
         # Frame ran to normal completion... check some invariants
-        if self.frames:  # pragma: no cover
-            raise PyVMError("Frames left over!")
-        if self.frame and self.frame.stack:  # pragma: no cover
-            raise PyVMError("Data left on stack! %r" % self.frame.stack)
+        if toplevel:
+            if self.frames:  # pragma: no cover
+                raise PyVMError("Frames left over!")
+            if self.frame and self.frame.stack:  # pragma: no cover
+                raise PyVMError("Data left on stack! %r" % self.frame.stack)
 
         return val
 
@@ -325,10 +358,12 @@ class PyVM(object):
             tb, value, exctype = self.popn(3)
             self.last_exception = exctype, value, tb
 
-    def parse_byte_and_args(self, byteCode):
-
+    def parse_byte_and_args(self, byte_code, replay=False):
         """ Parse 1 - 3 bytes of bytecode into
-        an instruction and optionally arguments."""
+        an instruction and optionally arguments.
+
+        Argument replay is used to handle breakpoints.
+        """
 
         f = self.frame
         f_code = f.f_code
@@ -338,88 +373,91 @@ class PyVM(object):
         # Note: There is never more than one argument.
         # The list size is used to indicate whether an argument
         # exists or not.
-        # FIMXE: remove and use intArg as a indicator of whether
+        # FIMXE: remove and use int_arg as a indicator of whether
         # the argument exists.
         arguments = []
-        intArg = None
+        int_arg = None
 
         while True:
             if f.fallthrough:
-                f.f_lasti = next_offset(byteCode, self.opc, f.f_lasti)
+                if not replay:
+                    f.f_lasti = next_offset(byte_code, self.opc, f.f_lasti)
             else:
                 # Jump instructions must set this False.
                 f.fallthrough = True
-            opoffset = f.f_lasti
-            line_number = self.frame.linestarts.get(opoffset, None)
+            offset = f.f_lasti
+            line_number = self.frame.linestarts.get(offset, None)
             if line_number is not None:
                 f.f_lineno = line_number
-            byteCode = byteint(co_code[opoffset])
-            byteName = self.opc.opname[byteCode]
-            arg_offset = opoffset + 1
+            if not replay:
+                byte_code = byteint(co_code[offset])
+            byte_name = self.opc.opname[byte_code]
+            arg_offset = offset + 1
             arg = None
 
-            if op_has_argument(byteCode, self.opc):
-                if PYTHON_VERSION >= 3.6:
-                    intArg = code2num(co_code, arg_offset) | extended_arg
+            if op_has_argument(byte_code, self.opc):
+                if self.version >= 3.6:
+                    int_arg = code2num(co_code, arg_offset) | extended_arg
                     # Note: Python 3.6.0a1 is 2, for 3.6.a3 and beyond we have 1
                     arg_offset += 1
-                    if byteCode == self.opc.EXTENDED_ARG:
-                        extended_arg = intArg << 8
+                    if byte_code == self.opc.EXTENDED_ARG:
+                        extended_arg = int_arg << 8
                         continue
                     else:
                         extended_arg = 0
                 else:
-                    intArg = (
+                    int_arg = (
                         code2num(co_code, arg_offset)
                         + code2num(co_code, arg_offset + 1) * 256
                         + extended_arg
                     )
                     arg_offset += 2
-                    if byteCode == self.opc.EXTENDED_ARG:
-                        extended_arg = intArg * 65536
+                    if byte_code == self.opc.EXTENDED_ARG:
+                        extended_arg = int_arg * 65536
                         continue
                     else:
                         extended_arg = 0
 
-                if byteCode in self.opc.CONST_OPS:
-                    arg = f_code.co_consts[intArg]
-                elif byteCode in self.opc.FREE_OPS:
-                    if intArg < len(f_code.co_cellvars):
-                        arg = f_code.co_cellvars[intArg]
+                if byte_code in self.opc.CONST_OPS:
+                    arg = f_code.co_consts[int_arg]
+                elif byte_code in self.opc.FREE_OPS:
+                    if int_arg < len(f_code.co_cellvars):
+                        arg = f_code.co_cellvars[int_arg]
                     else:
-                        var_idx = intArg - len(f.f_code.co_cellvars)
+                        var_idx = int_arg - len(f.f_code.co_cellvars)
                         arg = f_code.co_freevars[var_idx]
-                elif byteCode in self.opc.NAME_OPS:
-                    arg = f_code.co_names[intArg]
-                elif byteCode in self.opc.JREL_OPS:
+                elif byte_code in self.opc.NAME_OPS:
+                    arg = f_code.co_names[int_arg]
+                elif byte_code in self.opc.JREL_OPS:
                     # Many relative jumps are conditional,
                     # so setting f.fallthrough is wrong.
-                    arg = arg_offset + intArg
-                elif byteCode in self.opc.JABS_OPS:
+                    arg = arg_offset + int_arg
+                elif byte_code in self.opc.JABS_OPS:
                     # We probably could set fallthough, since many (all?)
                     # of these are unconditional, but we'll make the jump do
                     # the work of setting.
-                    arg = intArg
-                elif byteCode in self.opc.LOCAL_OPS:
-                    arg = f_code.co_varnames[intArg]
+                    arg = int_arg
+                elif byte_code in self.opc.LOCAL_OPS:
+                    arg = f_code.co_varnames[int_arg]
                 else:
-                    arg = intArg
+                    arg = int_arg
                 arguments = [arg]
             break
 
-        return byteName, byteCode, intArg, arguments, opoffset, line_number
+        return byte_name, byte_code, int_arg, arguments, offset, line_number
 
-    def log(self, byteName, intArg, arguments, opoffset, line_number):
+    def log(self, byte_name, int_arg, arguments, offset, line_number):
         """ Log arguments, block stack, and data stack for each opcode."""
         op = self.format_instruction(
             self.frame,
             self.opc,
-            byteName,
-            intArg,
+            byte_name,
+            int_arg,
             arguments,
-            opoffset,
+            offset,
             line_number,
             log.isEnabledFor(logging.DEBUG),
+            vm=self,
         )
         indent = "    " * (len(self.frames) - 1)
         stack_rep = repper(self.frame.stack)
@@ -429,25 +467,26 @@ class PyVM(object):
         log.debug("  %sblocks     : %s" % (indent, block_stack_rep))
         log.info("%s%s" % (indent, op))
 
-    def dispatch(self, byteName, intArg, arguments, opoffset, line_number):
-        """ Dispatch by bytename to the corresponding methods.
+    def dispatch(self, byte_name, int_arg, arguments, offset, line_number):
+        """ Dispatch by byte_name to the corresponding methods.
         Exceptions are caught and set on the virtual machine."""
 
         why = None
         self.in_exception_processing = False
+        byteop = self.byteop
         try:
-            if byteName.startswith("UNARY_"):
-                self.unaryOperator(byteName[6:])
-            elif byteName.startswith("BINARY_"):
-                self.binaryOperator(byteName[7:])
-            elif byteName.startswith("INPLACE_"):
-                self.inplaceOperator(byteName[8:])
-            elif "SLICE+" in byteName:
-                self.sliceOperator(byteName)
+            if byte_name.startswith("UNARY_"):
+                byteop.unaryOperator(byte_name[6:])
+            elif byte_name.startswith("BINARY_"):
+                byteop.binaryOperator(byte_name[7:])
+            elif byte_name.startswith("INPLACE_"):
+                byteop.inplaceOperator(byte_name[8:])
+            elif "SLICE+" in byte_name:
+                self.sliceOperator(byte_name)
             else:
                 # dispatch
-                if hasattr(self.byteop, byteName):
-                    bytecode_fn = getattr(self.byteop, byteName, None)
+                if hasattr(byteop, byte_name):
+                    bytecode_fn = getattr(byteop, byte_name, None)
                 if not bytecode_fn:  # pragma: no cover
                     raise PyVMError(
                         "Unknown bytecode type: %s\n\t%s"
@@ -455,14 +494,14 @@ class PyVM(object):
                             self.format_instruction(
                                 self.frame,
                                 self.opc,
-                                byteName,
-                                intArg,
+                                byte_name,
+                                int_arg,
                                 arguments,
-                                opoffset,
+                                offset,
                                 line_number,
                                 False,
                             ),
-                            byteName,
+                            byte_name,
                         )
                     )
                 why = bytecode_fn(*arguments)
@@ -481,16 +520,17 @@ class PyVM(object):
                             % self.format_instruction(
                                 self.frame,
                                 self.opc,
-                                byteName,
-                                intArg,
+                                byte_name,
+                                int_arg,
                                 arguments,
-                                opoffset,
+                                offset,
                                 line_number,
                                 False,
                             )
                         )
                     )
-                self.last_traceback = traceback_from_frame(self.frame)
+                if not self.last_traceback:
+                    self.last_traceback = traceback_from_frame(self.frame)
                 self.in_exception_processing = True
 
             why = "exception"
@@ -509,8 +549,9 @@ class PyVM(object):
             why = None
             return why
 
-        self.pop_block()
-        self.unwind_block(block)
+        if not (block.type == "except-handler" and why == "silenced"):
+            self.pop_block()
+            self.unwind_block(block)
 
         if block.type == "loop" and why == "break":
             why = None
@@ -554,48 +595,51 @@ class PyVM(object):
                 why = None
                 self.jump(block.handler)
                 return why
+            elif block.type == "except-handler" and why == "silenced":
+                # 3.5+ WITH_CLEANUP_FINISH
+                # Nothing needs to be done here.
+                return None
 
         return why
 
-    def run_frame(self, frame):
+    # Interpreter main loop
+    # This is analogous to CPython's _PyEval_EvalFramDefault() (in 3.x newer Python)
+    # or eval_frame() in older 2.x code.
+    def eval_frame(self, frame):
         """Run a frame until it returns (somehow).
 
         Exceptions are raised, the return value is returned.
 
         """
-        self.push_frame(frame)
-        self.f_code = self.frame.f_code
+        self.f_code = frame.f_code
         if frame.f_lasti == -1:
             # We were started new, not yielded back from.
             frame.f_lasti = 0
             # Don't increment before fetching next instruction.
             frame.fallthrough = False
-            byteCode = None
+            byte_code = None
         else:
-            byteCode = byteint(self.f_code.co_code[frame.f_lasti])
-            # byteCode == opcode["YIELD_VALUE"]?
+            byte_code = byteint(self.f_code.co_code[frame.f_lasti])
+            # byte_code == opcode["YIELD_VALUE"]?
 
-        # FIXME: we can use linestarts that is now located in the frame if this
-        # is a pyvmobj.Frame, and not a native frame.
-        self.frame.linestarts = dict(self.opc.findlinestarts(self.f_code, dup_lines=True))
-
-        opoffset = 0
+        self.push_frame(frame)
+        offset = 0
         while True:
 
             (
-                byteName,
-                byteCode,
-                intArg,
+                byte_name,
+                byte_code,
+                int_arg,
                 arguments,
-                opoffset,
+                offset,
                 line_number,
-            ) = self.parse_byte_and_args(byteCode)
+            ) = self.parse_byte_and_args(byte_code)
             if log.isEnabledFor(logging.INFO):
-                self.log(byteName, intArg, arguments, opoffset, line_number)
+                self.log(byte_name, int_arg, arguments, offset, line_number)
 
             # When unwinding the block stack, we need to keep track of why we
             # are doing it.
-            why = self.dispatch(byteName, intArg, arguments, opoffset, line_number)
+            why = self.dispatch(byte_name, int_arg, arguments, offset, line_number)
             if why == "exception":
                 # TODO: ceval calls PyTraceBack_Here, not sure what that does.
 
@@ -610,19 +654,20 @@ class PyVM(object):
                                 % self.format_instruction(
                                     frame,
                                     self.opc,
-                                    byteName,
-                                    intArg,
+                                    byte_name,
+                                    int_arg,
                                     arguments,
-                                    opoffset,
-                                line_number,
-                                False,
+                                    offset,
+                                    line_number,
+                                    False,
                                 )
                             )
                         )
-                    self.last_traceback = traceback_from_frame(self.frame)
+                    if self.last_traceback is None:
+                        self.last_traceback = traceback_from_frame(frame)
                     self.in_exception_processing = True
 
-            if why == "reraise":
+            elif why == "reraise":
                 why = "exception"
 
             if why != "yield":
@@ -638,8 +683,13 @@ class PyVM(object):
         self.pop_frame()
 
         if why == "exception":
-            if self.last_exception and self.last_exception[0]:
-                six.reraise(*self.last_exception)
+            last_exception = self.last_exception
+            if last_exception and last_exception[0]:
+                if isinstance(last_exception[2], Traceback):
+                    if not self.frame:
+                        raise PyVMUncaughtException(last_exception)
+                else:
+                    six.reraise(*self.last_exception)
             else:
                 raise PyVMError("Borked exception recording")
             # if self.exception and .... ?
@@ -651,75 +701,6 @@ class PyVM(object):
 
     ## Operators
 
-    UNARY_OPERATORS = {
-        "POSITIVE": operator.pos,
-        "NEGATIVE": operator.neg,
-        "NOT": operator.not_,
-        "CONVERT": repr,
-        "INVERT": operator.invert,
-    }
-
-    def unaryOperator(self, op):
-        x = self.pop()
-        self.push(self.UNARY_OPERATORS[op](x))
-
-    BINARY_OPERATORS = {
-        "POWER": pow,
-        "MULTIPLY": operator.mul,
-        "DIVIDE": getattr(operator, "div", lambda x, y: None),
-        "FLOOR_DIVIDE": operator.floordiv,
-        "TRUE_DIVIDE": operator.truediv,
-        "MODULO": operator.mod,
-        "ADD": operator.add,
-        "SUBTRACT": operator.sub,
-        "SUBSCR": operator.getitem,
-        "LSHIFT": operator.lshift,
-        "RSHIFT": operator.rshift,
-        "AND": operator.and_,
-        "XOR": operator.xor,
-        "OR": operator.or_,
-    }
-
-    if PYTHON_VERSION >= 3.5:
-        BINARY_OPERATORS["MATRIX_MULTIPLY"] = operator.matmul
-
-    def binaryOperator(self, op):
-        x, y = self.popn(2)
-        self.push(self.BINARY_OPERATORS[op](x, y))
-
-    def inplaceOperator(self, op):
-        x, y = self.popn(2)
-        if op == "POWER":
-            x **= y
-        elif op == "MULTIPLY":
-            x *= y
-        elif op in ["DIVIDE", "FLOOR_DIVIDE"]:
-            x //= y
-        elif op == "TRUE_DIVIDE":
-            x /= y
-        elif op == "MODULO":
-            x %= y
-        elif op == "ADD":
-            x += y
-        elif op == "SUBTRACT":
-            x -= y
-        elif op == "LSHIFT":
-            x <<= y
-        elif op == "RSHIFT":
-            x >>= y
-        elif op == "AND":
-            x &= y
-        elif op == "XOR":
-            x ^= y
-        elif op == "OR":
-            x |= y
-        # 3.5 on
-        elif op == "MATRIX_MULTIPLY":
-            operator.imatmul(x, y)
-        else:  # pragma: no cover
-            raise PyVMError("Unknown in-place operator: %r" % op)
-        self.push(x)
-
     def sliceOperator(self, op):
         start = 0
         end = None  # we will take this to mean end
